/*
 *	Sub-authentication Client
 *
 *	(c) 2017 Martin Mares <mj@ucw.cz>
 */

#include <ucw/lib.h>
#include <ucw/log.h>
#include <ucw/opt.h>
#include <ucw/strtonum.h>
#include <ucw-json/json.h>

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <time.h>
#include <unistd.h>

#include "autoconf.h"

/*** Arguments ***/

static char *socket_path = INSTALL_RUN_DIR "/subauthd/subauthd.socket";

static char *arg_user;
static char *arg_zone;
static char *arg_ident;
static char *arg_comment;
static int debug;
static char *arg_expire;
static int arg_auth_passwd;

/*** Communication with the daemon ***/

static struct json_context *js;
static struct json_node *rq;
static struct json_node *rp;

static void set_string(struct json_node *n, const char *key, const char *val)
{
  if (val)
    json_object_set(n, key, json_new_string(js, val));
}

static void set_uint(struct json_node *n, const char *key, uint val)
{
  if (val)
    json_object_set(n, key, json_new_number(js, val));
}

static struct json_node *get_child(struct json_node *obj, const char *key, uint type)
{
  struct json_node *n = json_object_get(obj, key);
  if (n && n->type != type)
    die("Malformed reply: %s has wrong type", key);
  return n;
}

static struct json_node *need_child(struct json_node *obj, const char *key, uint type)
{
  struct json_node *n = get_child(obj, key, type);
  if (!n)
    die("Malformed reply: missing %s", key);
  return n;
}

static void op_new(const char *op)
{
  js = json_new();
  rq = json_new_object(js);
  set_string(rq, "cmd", op);
  if (arg_user)
    set_string(rq, "login", arg_user);
  if (arg_auth_passwd)
    {
      char *auth_passwd = xstrdup(getpass("Current password: "));
      set_string(rq, "auth-passwd", auth_passwd);
    }
}

static void op_debug(const char *msg, struct json_node *n)
{
  if (!debug)
    return;

  struct fastbuf *out = bfdopen_shared(1, 4096);
  bprintf(out, ">>> %s\n", msg);
  js->format_options = JSON_FORMAT_INDENT;
  json_write(js, out, n);
  js->format_options = 0;
  bclose(out);
}

static void op_run(void)
{
  int sk = socket(PF_UNIX, SOCK_SEQPACKET, 0);
  if (sk < 0)
    die("socket(PF_UNIX, SOCK_SEQPACKET): %m");

  struct sockaddr_un sun;
  sun.sun_family = AF_UNIX;
  if (strlen(socket_path) >= sizeof(sun.sun_path))
    die("Socket path too long");
  strcpy(sun.sun_path, socket_path);

  if (connect(sk, (struct sockaddr *) &sun, sizeof(sun)) < 0)
    die("Cannot connect to %s: %m", socket_path);

  op_debug("Request", rq);

  struct fastbuf *rq_fb = fbgrow_create(4096);
  json_write(js, rq_fb, rq);
  byte *rq_buf;
  uint rq_len = fbgrow_get_buf(rq_fb, &rq_buf);
  if (send(sk, rq_buf, rq_len, 0) < 0)
    die("Cannot send request: %m");
  bclose(rq_fb);

  byte rp_buf[16384];
  int rp_len = recv(sk, rp_buf, sizeof(rp_buf), 0);
  if (rp_len < 0)
    die("Cannot receive reply: %m");

  struct fastbuf rp_fb;
  fbbuf_init_read(&rp_fb, rp_buf, rp_len, 0);
  rp = json_parse(js, &rp_fb);
  op_debug("Reply", rp);

  close(sk);

  if (rp->type != JSON_OBJECT)
    die("Malformed reply: Top-level node is not an object");
  struct json_node *err = need_child(rp, "error", JSON_STRING);
  if (strcmp(err->string, ""))
    die("Error: %s", err->string);
}

/*** Commands ***/

static void cmd_set_passwd(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("set-passwd");
  char *passwd = xstrdup(getpass("New password: "));
  char *again = xstrdup(getpass("Confirm password: "));
  if (strcmp(passwd, again))
    die("Passwords do not match");

  set_string(rq, "zone", arg_zone);
  set_string(rq, "passwd", passwd);

  op_run();
}

static void cmd_delete_passwd(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("delete-passwd");
  set_string(rq, "zone", arg_zone);

  op_run();
}

static void cmd_create_token(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("create-token");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "comment", arg_comment);

  op_run();
  struct json_node *jt = need_child(rp, "token", JSON_STRING);
  printf("%s\n", jt->string);
}

static void cmd_delete_token(void)
{
  if (!arg_zone || !arg_ident)
    opt_failure("--zone and --ident must be given");

  op_new("delete-token");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "ident", arg_ident);

  op_run();
}

static void cmd_change_token(void)
{
  if (!arg_zone || !arg_ident)
    opt_failure("--zone and --ident must be given");

  op_new("change-token");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "ident", arg_ident);
  set_string(rq, "comment", arg_comment);

  op_run();
}

static void cmd_create_acct(void)
{
  if (!arg_zone || !arg_user)
    opt_failure("--zone and --user must be given");

  op_new("create-acct");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "login", arg_user);

  op_run();
}

static void cmd_delete_acct(void)
{
  if (!arg_zone || !arg_user)
    opt_failure("--zone and --user must be given");

  op_new("delete-acct");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "login", arg_user);

  op_run();
}

static void cmd_delete_user(void)
{
  if (!arg_user)
    opt_failure("--user must be given");

  op_new("delete-acct");
  set_string(rq, "zone", "*");
  set_string(rq, "login", arg_user);

  op_run();
}

static void cmd_list(void)
{
  op_new("list-accts");
  op_run();

  struct json_node **jas = need_child(rp, "accounts", JSON_ARRAY)->elements;
  if (!GARY_SIZE(jas))
    {
      puts("No accounts defined");
      return;
    }

  printf("%-16s %-6s %-5s %-19s %-8s  %s\n",
    "Zone", "Type", "Ident", "Last modified", "PassAuth", "Comment");
  for (uint i=0; i < GARY_SIZE(jas); i++)
    {
      struct json_node *ja = jas[i];
      struct json_node *jz = need_child(ja, "zone", JSON_STRING);
      struct json_node **jts = need_child(ja, "tokens", JSON_ARRAY)->elements;
      struct json_node *jpassauth = get_child(ja, "allow-passwd-auth", JSON_NUMBER);

      if (!GARY_SIZE(jts))
	{
	  printf("%-16s (no tokens defined%s)\n",
	    jz->string,
	    jpassauth ? (jpassauth->number ? ", password auth allowed" : ", password auth disabled") : "");
	  continue;
	}

      for (uint j=0; j < GARY_SIZE(jts); j++)
	{
	  struct json_node *lt = jts[j];
	  struct json_node *jtype = need_child(lt, "type", JSON_STRING);
	  struct json_node *jident = get_child(lt, "ident", JSON_STRING);
	  struct json_node *jcomment = get_child(lt, "comment", JSON_STRING);
	  struct json_node *jlastmod = need_child(lt, "lastmod", JSON_NUMBER);
	  time_t lm = jlastmod->number;
	  struct tm *tm = localtime(&lm);
	  printf("%-16s %-6s %-5s %04d-%02d-%02d %02d:%02d:%02d %8s  %s\n",
	    jz->string,
	    jtype->string,
	    (jident ? jident->string : "-"),
	    tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
	    tm->tm_hour, tm->tm_min, tm->tm_sec,
	    (jpassauth && jpassauth->number ? "yes" : "-"),
	    (jcomment ? jcomment->string : ""));
	}
    }
}

static void cmd_zones(void)
{
  op_new("list-zones");
  op_run();

  struct json_node **jzs = need_child(rp, "zones", JSON_ARRAY)->elements;
  printf("%-16s %6s %6s %6s %8s  %s\n", "Zone", "Passwd", "Tokens", "MaxTmp", "PassAuth", "Description");
  for (uint i=0; i < GARY_SIZE(jzs); i++)
    {
      struct json_node *jz = jzs[i];
      struct json_node *jname = need_child(jz, "name", JSON_STRING);
      struct json_node *jdesc = get_child(jz, "desc", JSON_STRING);
      struct json_node *jpass = get_child(jz, "allow-passwd", JSON_NUMBER);
      struct json_node *jtokens = get_child(jz, "allow-tokens", JSON_NUMBER);
      struct json_node *jpassauth = get_child(jz, "allow-passwd-auth", JSON_NUMBER);
      struct json_node *jtemp = get_child(jz, "max-temp-validity", JSON_NUMBER);
      printf("%-16s %6s %6d %6d %8s  %s\n",
	jname->string,
	(jpass && jpass->number ? "yes" : "-"),
	(jtokens ? (uint) jtokens->number : 0),
	(jtemp ? (uint) jtemp->number : 0),
	(jpassauth && jpassauth->number ? "yes" : "-"),
	(jdesc ? jdesc->string : "(no description)"));
    }
}

static void cmd_login(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  if (!arg_user)
    {
      uid_t uid = getuid();
      struct passwd *pw = getpwuid(uid);
      if (!pw)
	die("You do not exist");
      arg_user = xstrdup(pw->pw_name);
    }

  char *passwd = xstrdup(getpass("Password: "));

  op_new("login");
  set_string(rq, "zone", arg_zone);
  set_string(rq, "passwd", passwd);

  op_run();
  printf("OK\n");
}

static uint parse_expire(void)
{
  if (!arg_expire)
    arg_expire = "5m";

  uint expire;
  const char *next;
  const char *err = str_to_uint(&expire, arg_expire, &next, 10);
  if (err || !next[0] || next[1])
    opt_failure("Invalid syntax of --expire");

  switch (next[0])
    {
    case 'h':
      return expire * 3600;
    case 'm':
      return expire * 60;
    case 's':
      return expire;
    default:
      opt_failure("--expire uses an unknown unit");
    }
}

static void cmd_temp_token(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("create-temp");
  set_string(rq, "zone", arg_zone);
  set_uint(rq, "validity", parse_expire());

  op_run();
  struct json_node *jt = need_child(rp, "token", JSON_STRING);
  printf("%s\n", jt->string);
}

static void cmd_allow_passwd_auth(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("allow-passwd-auth");
  set_string(rq, "zone", arg_zone);
  set_uint(rq, "allow", 1);

  op_run();
}

static void cmd_disallow_passwd_auth(void)
{
  if (!arg_zone)
    opt_failure("--zone must be given");

  op_new("allow-passwd-auth");
  set_string(rq, "zone", arg_zone);
  json_object_set(rq, "allow", json_new_number(js, 0));

  op_run();
}

static void cmd_raw(void)
{
  op_new("");

  struct fastbuf *in = bfdopen_shared(0, 4096);
  rq = json_parse(js, in);
  bclose(in);

  op_run();
}

enum command {
  CMD_SET_PASSWD,
  CMD_DELETE_PASSWD,
  CMD_CREATE_TOKEN,
  CMD_DELETE_TOKEN,
  CMD_CHANGE_TOKEN,
  CMD_CREATE_ACCT,
  CMD_DELETE_ACCT,
  CMD_DELETE_USER,
  CMD_LIST,
  CMD_ZONES,
  CMD_LOGIN,
  CMD_TEMP_TOKEN,
  CMD_ALLOW_PASSWD_AUTH,
  CMD_DISALLOW_PASSWD_AUTH,
  CMD_RAW,
  CMD_MAX
};

void (* const command_handlers[CMD_MAX])(void) = {
  [CMD_SET_PASSWD] = cmd_set_passwd,
  [CMD_DELETE_PASSWD] = cmd_delete_passwd,
  [CMD_CREATE_TOKEN] = cmd_create_token,
  [CMD_DELETE_TOKEN] = cmd_delete_token,
  [CMD_CHANGE_TOKEN] = cmd_change_token,
  [CMD_CREATE_ACCT] = cmd_create_acct,
  [CMD_DELETE_ACCT] = cmd_delete_acct,
  [CMD_DELETE_USER] = cmd_delete_user,
  [CMD_LIST] = cmd_list,
  [CMD_ZONES] = cmd_zones,
  [CMD_LOGIN] = cmd_login,
  [CMD_TEMP_TOKEN] = cmd_temp_token,
  [CMD_ALLOW_PASSWD_AUTH] = cmd_allow_passwd_auth,
  [CMD_DISALLOW_PASSWD_AUTH] = cmd_disallow_passwd_auth,
  [CMD_RAW] = cmd_raw,
};

static int command = -1;

/*** Main ***/

static const struct opt_section options = {
  OPT_ITEMS {
    OPT_HELP("A client to the sub-authentication daemon."),
    OPT_HELP("Usage: subauth [general options] <command> [options]"),
    OPT_HELP(""),
    OPT_HELP("General options:"),
    OPT_STRING('s', "socket", socket_path, OPT_REQUIRED_VALUE, "path\tPath to daemon control socket"),
    OPT_BOOL(0, "debug", debug, 0, "\tDump raw communication with the daemon"),
    OPT_HELP_OPTION,
    OPT_HELP(""),
    OPT_HELP("User commands:"),
    OPT_SWITCH('l', "list",			command, CMD_LIST,			OPT_SINGLE, "\tList tokens for all zones"),
    OPT_SWITCH(0,   "zones",			command, CMD_ZONES,			OPT_SINGLE, "\tList all known zones"),
    OPT_SWITCH(0,   "passwd",			command, CMD_SET_PASSWD,		OPT_SINGLE, "\tSet password"),
    OPT_SWITCH(0,   "delete-passwd",		command, CMD_DELETE_PASSWD,		OPT_SINGLE, "\tRemove password"),
    OPT_SWITCH(0,   "create-token",		command, CMD_CREATE_TOKEN,		OPT_SINGLE, "\tCreate a new token"),
    OPT_SWITCH(0,   "delete-token",		command, CMD_DELETE_TOKEN,		OPT_SINGLE, "\tRemove an existing token"),
    OPT_SWITCH(0,   "change-token",		command, CMD_CHANGE_TOKEN,		OPT_SINGLE, "\tChange the comment of an existing token"),
    OPT_SWITCH(0,   "login",			command, CMD_LOGIN,			OPT_SINGLE, "\tTry to log in"),
    OPT_SWITCH(0,   "temp-token",		command, CMD_TEMP_TOKEN,		OPT_SINGLE, "\tCreate a temporary token"),
    OPT_SWITCH(0,   "allow-passwd-auth",	command, CMD_ALLOW_PASSWD_AUTH,		OPT_SINGLE, "\tAllow management using password authentication"),
    OPT_SWITCH(0,   "disallow-passwd-auth",	command, CMD_DISALLOW_PASSWD_AUTH,	OPT_SINGLE, "\tDisallow account management"),
    OPT_HELP(""),
    OPT_HELP("Administrative commands:"),
    OPT_SWITCH(0,   "create-acct",	command, CMD_CREATE_ACCT,	OPT_SINGLE, "\tCreate an account"),
    OPT_SWITCH(0,   "delete-acct",	command, CMD_DELETE_ACCT,	OPT_SINGLE, "\tDelete an account"),
    OPT_SWITCH(0,   "delete-user",	command, CMD_DELETE_USER,	OPT_SINGLE, "\tDelete a user with all his accounts"),
    OPT_SWITCH(0,   "raw",		command, CMD_RAW,		OPT_SINGLE, "\tSend raw JSON message to the daemon"),
    OPT_HELP(""),
    OPT_HELP("Command options:"),
    OPT_STRING('u', "user", arg_user, OPT_REQUIRED_VALUE, "login\tUser to act on (default: whoever calls it)"),
    OPT_STRING('z', "zone", arg_zone, OPT_REQUIRED_VALUE, "zone\tAuthentication zone"),
    OPT_STRING('i', "ident", arg_ident, OPT_REQUIRED_VALUE, "id\tToken ID ('*' for all tokens in zone)"),
    OPT_STRING('c', "comment", arg_comment, OPT_REQUIRED_VALUE, "string\tHuman-readable description of token"),
    OPT_STRING('x', "expire", arg_expire, OPT_REQUIRED_VALUE, "n(h|m|s)\tSet expiration time of a temporary token (default: 5m)"),
    OPT_BOOL(0, "auth-passwd", arg_auth_passwd, 0, "\tAsk for current password to authorize the operation"),
    OPT_END
  }
};

int main(int argc UNUSED, char **argv)
{
  log_set_format(log_default_stream(), 0, 0);
  opt_parse(&options, argv+1);

  if (command < 0)
    opt_failure("No command given");
  if (!command_handlers[command])
    opt_failure("Command not implemented");
  command_handlers[command]();

  return 0;
}
