/*
 *	Sub-authentication Daemon: Authentication database
 *
 *	(c) 2017 Martin Mares <mj@ucw.cz>
 */

#include <ucw/lib.h>
#include <ucw/base64.h>
#include <ucw/gary.h>
#include <ucw/stkstring.h>
#include <ucw/string.h>
#include <ucw/unaligned.h>

#include <errno.h>
#include <fcntl.h>
#include <gcrypt.h>
#include <time.h>
#include <unistd.h>

#include "subauthd.h"

struct auth_token *auth_fake_token;

#define HASH_NODE struct auth_user
#define HASH_PREFIX(x) user_hash_##x
#define HASH_KEY_ENDSTRING login
#define HASH_WANT_FIND
#define HASH_WANT_LOOKUP
#define HASH_WANT_REMOVE
#define HASH_GIVE_INIT_DATA

static void user_hash_init_data(struct auth_user *au)
{
  clist_init(&au->accounts);
}

#include <ucw/hashtable.h>

struct auth_zone *auth_find_zone(const char *name)
{
  CLIST_FOR_EACH(struct auth_zone *, z, zone_list)
    if (!strcmp(z->name, name))
      return z;
  return NULL;
}

struct auth_user *auth_find_user(const char *login, bool create)
{
  if (create)
    return user_hash_lookup((char *) login);
  else
    return user_hash_find((char *) login);
}

struct auth_acct *auth_find_acct(struct auth_user *au, struct auth_zone *az, bool create)
{
  CLIST_FOR_EACH(struct auth_acct *, aa, au->accounts)
    if (aa->zone == az)
      return aa;

  if (!create)
    return NULL;

  struct auth_acct *aa = xmalloc_zero(sizeof(*aa));
  clist_add_tail(&au->accounts, &aa->n);
  aa->user = au;
  aa->zone = az;
  clist_init(&aa->tokens);
  return aa;
}

struct auth_token *auth_find_token_passwd(struct auth_acct *aa)
{
  CLIST_FOR_EACH(struct auth_token *, at, aa->tokens)
    if (at->type == TOKEN_PASSWORD)
      return at;
  return NULL;
}

struct auth_token *auth_find_token_generated(struct auth_acct *aa, const char *ident)
{
  CLIST_FOR_EACH(struct auth_token *, at, aa->tokens)
    if (at->type == TOKEN_GENERATED && !strcmp(at->ident, ident))
      return at;
  return NULL;
}

void auth_delete_user(struct auth_user *au)
{
  struct auth_acct *aa;
  while (aa = clist_head(&au->accounts))
    auth_delete_acct(aa);

  user_hash_remove(au);
}

void auth_delete_acct(struct auth_acct *ac)
{
  struct auth_token *at;
  while (at = clist_head(&ac->tokens))
    auth_delete_token(at);
  clist_remove(&ac->n);
}

void auth_delete_token(struct auth_token *at)
{
  clist_remove(&at->n);
  xfree(at->salt);
  xfree(at->hash);
  xfree(at->comment);
}

struct auth_token *auth_create_token(struct auth_acct *aa)
{
  struct auth_token *at = xmalloc_zero(sizeof(*at));
  clist_add_tail(&aa->tokens, &at->n);
  at->acct = aa;
  at->last_modified = time(NULL);
  return at;
}

static void auth_hash(char *hash_text, struct auth_token *at, const char *passwd)
{
  // This is PBKDF2 based on SHA-256-HMAC

  gcry_md_hd_t md;
  if (gcry_md_open(&md, GCRY_MD_SHA256, GCRY_MD_FLAG_HMAC) != GPG_ERR_NO_ERROR)
    die("gcry_md_open(CGRY_MD_SHA256) failed");
  if (gcry_md_setkey(md, passwd, strlen(passwd)) != GPG_ERR_NO_ERROR)
    die("gcry_md_setkey() failed");

  byte current_hash[HASH_BYTES];
  gcry_md_write(md, at->salt, strlen(at->salt));
  byte zeroes[4] = { 0, 0, 0, 0 };
  gcry_md_write(md, zeroes, 4);
  memcpy(current_hash, gcry_md_read(md, 0), HASH_BYTES);

  byte accumulator[HASH_BYTES];
  memcpy(accumulator, current_hash, HASH_BYTES);

  for (uint i=1; i < at->iterations; i++)
    {
      gcry_md_reset(md);
      gcry_md_write(md, current_hash, HASH_BYTES);
      memcpy(current_hash, gcry_md_read(md, 0), HASH_BYTES);
      for (uint j=0; j<HASH_BYTES; j++)
	accumulator[j] ^= current_hash[j];
    }

  gcry_md_close(md);

  uint out_len = base64_encode(hash_text, accumulator, HASH_BYTES);
  hash_text[out_len] = 0;
}

static void set_salt(struct auth_token *at)
{
  byte salt_bytes[DEFAULT_SALT_BYTES];
  gcry_randomize(salt_bytes, DEFAULT_SALT_BYTES, GCRY_STRONG_RANDOM);

  char salt_text[2*DEFAULT_SALT_BYTES + 1];
  uint salt_len = base64_encode(salt_text, salt_bytes, DEFAULT_SALT_BYTES);
  salt_text[salt_len] = 0;

  at->salt = xstrdup(salt_text);
  at->iterations = DEFAULT_HASH_ITERATIONS;
}

void auth_set_token_passwd(struct auth_token *at, const char *passwd)
{
  at->type = TOKEN_PASSWORD;
  set_salt(at);

  char hash[MAX_TEXT_HASH_SIZE];
  auth_hash(hash, at, passwd);
  at->hash = xstrdup(hash);
}

char *auth_set_token_generated(struct auth_token *at, const char *comment, struct mempool *pool)
{
  char ident_text[2*DEFAULT_IDENT_BYTES + 1];
  do
    {
      byte ident_bytes[DEFAULT_IDENT_BYTES];
      gcry_randomize(ident_bytes, DEFAULT_IDENT_BYTES, GCRY_STRONG_RANDOM);
      mem_to_hex(ident_text, ident_bytes, DEFAULT_IDENT_BYTES, 0);
    }
  while (auth_find_token_generated(at->acct, ident_text));

  at->type = TOKEN_GENERATED;
  at->ident = xstrdup(ident_text);
  at->comment = xstrdup(comment);
  set_salt(at);

  byte token_bytes[DEFAULT_GENERATED_BYTES];
  gcry_randomize(token_bytes, DEFAULT_GENERATED_BYTES, GCRY_STRONG_RANDOM);

  uint token_len = 2*DEFAULT_IDENT_BYTES + 1 + 2*DEFAULT_GENERATED_BYTES + 1;
  char *token_text = mp_alloc(pool, token_len);
  memcpy(token_text, ident_text, 2*DEFAULT_IDENT_BYTES);
  char *token_rhs = token_text + 2*DEFAULT_IDENT_BYTES;
  *token_rhs++ = '-';
  uint out_len = base64_encode(token_rhs, token_bytes, DEFAULT_GENERATED_BYTES);
  if (out_len > 0 && token_rhs[out_len-1] == '=')
    out_len--;
  token_rhs[out_len] = 0;

  char hash[MAX_TEXT_HASH_SIZE];
  auth_hash(hash, at, token_rhs);
  at->hash = xstrdup(hash);

  return token_text;
}

bool auth_check_token(struct auth_token *at, const char *passwd)
{
  char hash[MAX_TEXT_HASH_SIZE];
  auth_hash(hash, at, passwd);
  return !strcmp(at->hash, hash);
}

static void auth_create_fake_token(void)
{
  struct auth_token *at = xmalloc_zero(sizeof(*at));
  auth_set_token_passwd(at, "TheKeyIsUnderTheFlyingCarpet");
  auth_fake_token = at;
}

static void NONRET db_corrupted(const char *reason)
{
  die("Database file corrupted: %s", reason);
}

static void db_parse_user(struct json_node *ju)
{
  if (ju->type != JSON_OBJECT)
    db_corrupted("Expected object");

  const char *login = get_string(ju, "l");
  if (!login)
    db_corrupted("User has no login");

  struct json_node **jas = get_array(ju, "a");
  if (!jas)
    db_corrupted("User has no accounts");

  struct auth_user *au = auth_find_user(login, 1);
  if (clist_head(&au->accounts))
    db_corrupted("Multiple users with the same login");

  for (uint i=0; i < GARY_SIZE(jas); i++)
    {
      struct json_node *ja = jas[i];

      const char *zone_name = get_string(ja, "z");
      if (!zone_name)
	db_corrupted("Account has no zone");

      struct auth_acct *aa = xmalloc_zero(sizeof(*aa));
      clist_add_tail(&au->accounts, &aa->n);
      aa->user = au;
      aa->zone = auth_find_zone(zone_name);
      if (!aa->zone)
	die("Database defines accounts in zone %s, which is not configured", zone_name);
      clist_init(&aa->tokens);

      uint allow_passwd_auth;
      if (!get_uint(ja, "p", &allow_passwd_auth))
	allow_passwd_auth = 0;
      aa->allow_passwd_auth = allow_passwd_auth;

      struct json_node **jts = get_array(ja, "t");
      if (jts)
	{
	  for (uint j=0; j < GARY_SIZE(jts); j++)
	    {
	      struct json_node *jt = jts[j];
	      if (jt->type != JSON_OBJECT)
		db_corrupted("Token is not an object");

	      uint type;
	      if (!get_uint(jt, "t", &type) || type >= TOKEN_NUM_TYPES)
		db_corrupted("Token has no valid type");

	      const char *salt = get_string(jt, "s");
	      const char *hash = get_string(jt, "h");
	      const char *cmt = get_string(jt, "c");
	      const char *ident = get_string(jt, "i");
	      if (!salt || !hash)
		db_corrupted("Token must have salt and hash");

	      struct auth_token *at = auth_create_token(aa);
	      at->type = type;
	      at->salt = xstrdup(salt);
	      at->hash = xstrdup(hash);
	      at->comment = xstrdup(cmt);
	      at->ident = xstrdup(ident);

	      uint lastmod;
	      if (!get_uint(jt, "m", &lastmod))
		db_corrupted("Token has invalid last modified time");
	      at->last_modified = lastmod;

	      if (!get_uint(jt, "n", &at->iterations))
		db_corrupted("Token has invalid number of hash iterations");
	    }
	}
    }
}

void auth_change_token_comment(struct auth_token *at, const char *comment)
{
  xfree(at->comment);
  at->comment = xstrdup(comment);
}

static void db_read(void)
{
  struct fastbuf *fb = bopen_try(database_name, O_RDONLY, 65536);
  if (!fb)
    {
      msg(L_WARN, "No database file found, starting with no users");
      return;
    }

  struct json_context *js = json_new();
  json_set_input(js, fb);

  struct json_node *sep = json_next_token(js);
  if (sep->type != JSON_BEGIN_ARRAY)
    db_corrupted("Does not start with '['");

  for (;;)
    {
      json_push(js);

      sep = json_peek_token(js);
      if (sep->type == JSON_END_ARRAY)
	{
	  json_next_token(js);
	  json_pop(js);
	  break;
	}
      else if (sep->type == JSON_VALUE_SEP)
	json_next_token(js);

      struct json_node *ju = json_next_value(js);
      db_parse_user(ju);
      json_pop(js);
    }

  sep = json_next_token(js);
  if (sep->type != JSON_EOF)
    db_corrupted("Garbage at the end");

  json_delete(js);
  bclose(fb);
}

void db_write(void)
{
  char *temp_name = stk_strcat(database_name, ".new");
  struct fastbuf *fb = bopen(temp_name, O_WRONLY | O_CREAT | O_TRUNC, 65536);
  struct json_context *js = json_new();
  json_set_output(js, fb);
  bputs(fb, "[\n");
  uint cnt = 0;

  HASH_FOR_ALL(user_hash, au)
    {
      json_push(js);
      if (cnt)
	bputs(fb, ",\n");
      struct json_node *ju = json_new_object(js);
      json_object_set(ju, "l", json_new_string_ref(js, au->login));
      struct json_node *jas = json_new_array(js);
      json_object_set(ju, "a", jas);
      CLIST_FOR_EACH(struct auth_acct *, aa, au->accounts)
	{
	  struct json_node *ja = json_new_object(js);
	  json_array_append(jas, ja);
	  json_object_set(ja, "z", json_new_string_ref(js, aa->zone->name));
	  json_object_set(ja, "p", json_new_number(js, aa->allow_passwd_auth));
	  struct json_node *jts = json_new_array(js);
	  CLIST_FOR_EACH(struct auth_token *, at, aa->tokens)
	    {
	      struct json_node *jt = json_new_object(js);
	      json_array_append(jts, jt);
	      json_object_set(jt, "t", json_new_number(js, at->type));
	      if (at->salt)
		json_object_set(jt, "s", json_new_string_ref(js, at->salt));
	      if (at->hash)
		json_object_set(jt, "h", json_new_string_ref(js, at->hash));
	      if (at->comment)
		json_object_set(jt, "c", json_new_string_ref(js, at->comment));
	      if (at->ident)
		json_object_set(jt, "i", json_new_string_ref(js, at->ident));
	      json_object_set(jt, "n", json_new_number(js, at->iterations));
	      json_object_set(jt, "m", json_new_number(js, at->last_modified));
	    }
	  if (GARY_SIZE(jts->elements))
	    json_object_set(ja, "t", jts);
	}
      json_write_value(js, ju);
      json_pop(js);
      cnt++;
    }
  HASH_END_FOR;

  bputs(fb, "\n]\n");
  json_delete(js);
  bclose(fb);

  char *old_name = stk_strcat(database_name, ".old");
  if (rename(database_name, old_name) < 0 && errno != ENOENT)
    die("Cannot rename %s to %s: %m", database_name, old_name);

  if (rename(temp_name, database_name) < 0)
    die("Cannot rename %s to %s: %m", temp_name, database_name);
}

void auth_init(void)
{
  if (!gcry_check_version(GCRYPT_VERSION))
    die("libgcrypt version mismatch");
  gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
  gcry_control(GCRYCTL_INITIALIZATION_FINISHED, 0);

  user_hash_init();
  db_read();
  auth_create_fake_token();
}
