/*
 *	Sub-authentication Daemon
 *
 *	(c) 2017 Martin Mares <mj@ucw.cz>
 */

#undef LOCAL_DEBUG

#include <ucw/lib.h>
#include <ucw/conf.h>
#include <ucw/log.h>
#include <ucw/opt.h>
#include <ucw/trans.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>

#include "subauthd.h"

// Configuration
static char *socket_path = "subauthd.socket";
static uint max_connections = ~0U;
clist zone_list;
char *database_name = "subauthd.db";
char *temp_key_file;
char *log_stream_name;
static uint max_packet_size = 16384;
uint max_comment_size = 100;

static struct main_file listen_socket;
static uint num_connections;

static byte *packet_buffer;
static byte oob_data_buffer[MAX_OOB_DATA_SIZE];

static int socket_read_handler(struct main_file *fi);
static int socket_write_handler(struct main_file *fi);

static void client_close(struct client *c)
{
  DBG("Closing connection");
  file_del(&c->socket);
  timer_del(&c->timer);
  close(c->socket.fd);
  json_delete(c->json);
  mp_delete(c->pool);		// This includes the connection structure
  num_connections--;
}

static void socket_timeout_handler(struct main_timer *tm)
{
  struct client *c = tm->data;
  msg(L_INFO, "Client timeout");
  client_close(c);
}

static void try_send_reply(struct client *c)
{
  int len;

  TRANS_TRY
    {
      struct fastbuf fb;
      fbbuf_init_write(&fb, packet_buffer, max_packet_size);
      fb_tie(&fb);
      json_write(c->json, &fb, c->reply);
      len = fbbuf_count_written(&fb);
    }
  TRANS_CATCH(x)
    {
      msg(L_ERROR, "Unable to construct reply, it is probably too long");
      struct fastbuf fb2;
      fbbuf_init_write(&fb2, packet_buffer, max_packet_size);
      bputs(&fb2, "{ \"error\": \"Reply too long\" }\n");
      len = fbbuf_count_written(&fb2);
    }
  TRANS_END;

  DBG("Sending reply of %d bytes", len);
  if (send(c->socket.fd, packet_buffer, len, 0) < 0)
    {
      if (errno == EAGAIN || errno == EINTR)
	{
	  DBG("Postponed send");
	  c->socket.write_handler = socket_write_handler;
	  file_chg(&c->socket);
	}
      msg(L_ERROR, "Client write error: %m");
      client_close(c);
    }
  else
    {
      DBG("Reply sent");
      c->socket.read_handler = socket_read_handler;
      c->socket.write_handler = NULL;
      file_chg(&c->socket);
      timer_add_rel(&c->timer, SOCKET_TIMEOUT);
    }
}

static void send_reply(struct client *c)
{
  timer_add_rel(&c->timer, SOCKET_TIMEOUT);
  try_send_reply(c);
}

static void received_packet(struct client *c, byte *pkt, int len)
{
  json_reset(c->json);

  struct fastbuf fb;
  fbbuf_init_read(&fb, pkt, len, 0);

  c->reply = json_new_object(c->json);

  TRANS_TRY
    {
      c->request = json_parse(c->json, &fb);
    }
  TRANS_CATCH(x)
    {
      json_object_set(c->reply, "error", json_new_string_ref(c->json, "Parse error"));
      send_reply(c);
      return;
    }
  TRANS_END;

  cmd_dispatch(c);
  send_reply(c);
}

static int socket_write_handler(struct main_file *fi)
{
  struct client *c = fi->data;
  try_send_reply(c);
  return HOOK_IDLE;
}

static int socket_read_handler(struct main_file *fi)
{
  struct client *c = fi->data;

  struct iovec iov = {
    .iov_base = packet_buffer,
    .iov_len = max_packet_size,
  };

  struct msghdr mh = {
    .msg_iov = &iov,
    .msg_iovlen = 1,
    .msg_control = oob_data_buffer,
    .msg_controllen = MAX_OOB_DATA_SIZE,
  };

  ssize_t len = recvmsg(fi->fd, &mh, 0);
  if (len < 0)
    {
      if (errno != EAGAIN && errno != EINTR)
	msg(L_ERROR, "Socket read: %m");
      return HOOK_IDLE;
    }

  if (!len)
    {
      client_close(c);
      return HOOK_IDLE;
    }

  struct ucred *cred = NULL;
  for (struct cmsghdr *cm = CMSG_FIRSTHDR(&mh); cm; cm = CMSG_NXTHDR(&mh, cm))
    {
      if (cm->cmsg_level == SOL_SOCKET)
	{
	  if (cm->cmsg_type == SCM_RIGHTS)
	    {
	      // We are not interested in receiving file descriptor, but despite
	      // that they could be attached to the message. If it happens, simply
	      // close them.
	      int *fdptr = (int *) CMSG_DATA(cm);
	      int nfd = cm->cmsg_len / sizeof(int);
	      for (int i=0; i<nfd; i++)
		close(fdptr[i]);
	    }
	  else if (cm->cmsg_type == SCM_CREDENTIALS)
	    {
	      ASSERT(cm->cmsg_len >= sizeof(cred));
	      cred = (struct ucred *) CMSG_DATA(cm);
	    }
	}
    }

  if (!cred)
    {
      msg(L_ERROR, "Dropping message with no credentials");
      return HOOK_RETRY;
    }

  DBG("Got message from UID %d", (int) cred->uid);
  c->uid = cred->uid;

  fi->read_handler = NULL;
  file_chg(fi);

  received_packet(c, packet_buffer, len);

  // This has to be HOOK_IDLE, since the connection may have been already gone.
  return HOOK_IDLE;
}

static int listen_read_handler(struct main_file *fi)
{
  struct sockaddr_un client;
  socklen_t addr_len = sizeof(client);

  int new_sk = accept(fi->fd, &client, &addr_len);
  if (new_sk < 0)
    {
      if (errno != EAGAIN && errno != EINTR)
	msg(L_ERROR, "Socket accept: %m");
      return HOOK_IDLE;
    }

  if (num_connections >= max_connections)
    {
      msg(L_WARN, "Too many connections (you might need to increase MaxConnections)");
      close(new_sk);
      return HOOK_IDLE;
    }
  num_connections++;

  DBG("Accepted connection");

  if (fcntl(new_sk, F_SETFL, fcntl(new_sk, F_GETFL) | O_NONBLOCK) < 0)
    die("Cannot set O_NONBLOCK: %m");

  struct mempool *mp = mp_new(4096);
  struct client *c = mp_alloc_zero(mp, sizeof(*c));
  c->pool = mp;
  c->json = json_new();

  c->socket.fd = new_sk;
  c->socket.read_handler = socket_read_handler;
  c->socket.data = c;
  file_add(&c->socket);

  c->timer.handler = socket_timeout_handler;
  c->timer.data = c;
  timer_add_rel(&c->timer, SOCKET_TIMEOUT);

  return HOOK_RETRY;
}

static void init_socket(void)
{
  packet_buffer = xmalloc(max_packet_size);

  int sk = socket(PF_UNIX, SOCK_SEQPACKET, 0);
  if (sk < 0)
    die("socket(PF_UNIX, SOCK_SEQPACKET): %m");

  if (fcntl(sk, F_SETFL, fcntl(sk, F_GETFL) | O_NONBLOCK) < 0)
    die("Cannot set O_NONBLOCK: %m");

  struct sockaddr_un sun;
  sun.sun_family = AF_UNIX;
  if (strlen(socket_path) >= sizeof(sun.sun_path))
    die("SocketPath too long");
  strcpy(sun.sun_path, socket_path);

  if (unlink(socket_path) < 0 && errno != ENOENT)
    die("Cannot unlink old socket %s: %m", socket_path);

  if (bind(sk, (struct sockaddr *) &sun, sizeof(sun)) < 0)
    die("Cannot bind to %s: %m", socket_path);

  if (listen(sk, 64) < 0)
    die("listen(): %m");

  int one = 1;
  if (setsockopt(sk, SOL_SOCKET, SO_PASSCRED, &one, sizeof(one)) < 0)
    die("setsockopt(SO_PASSCRED): %m");

  listen_socket.fd = sk;
  listen_socket.read_handler = listen_read_handler;
  file_add(&listen_socket);

  if (chmod(socket_path, 0666) < 0)
    die("Cannot chmod socket: %m");

  msg(L_INFO, "Listening on %s", socket_path);
}

static void sig_term(struct main_signal *ms UNUSED)
{
  msg(L_INFO, "Shutting down");
  main_shut_down();
}

static void init_signals(void)
{
  static struct main_signal ms_sigterm = {
    .signum = SIGTERM,
    .handler = sig_term,
  };
  signal_add(&ms_sigterm);
}

static char *zone_commit(void *z_)
{
  struct auth_zone *z = z_;
  if (!z->name)
    return "A zone must have a name";
  return NULL;
}

static struct cf_section zone_config = {
  CF_TYPE(struct auth_zone),
  CF_COMMIT(zone_commit),
  CF_ITEMS {
    CF_STRING("Name", PTR_TO(struct auth_zone, name)),
    CF_STRING("Description", PTR_TO(struct auth_zone, desc)),
    CF_UINT("AutoCreateAcct", PTR_TO(struct auth_zone, auto_create_acct)),
    CF_UINT("AllowPasswd", PTR_TO(struct auth_zone, allow_passwd)),
    CF_UINT("AllowPasswdAuth", PTR_TO(struct auth_zone, allow_passwd_auth)),
    CF_UINT("AllowTokens", PTR_TO(struct auth_zone, allow_tokens)),
    CF_UINT("MaxTempValidity", PTR_TO(struct auth_zone, max_temp_validity)),
    CF_END
  }
};

static struct cf_section daemon_config = {
  CF_ITEMS {
    CF_STRING("SocketPath", &socket_path),
    CF_UINT("MaxConnections", &max_connections),
    CF_UINT("MaxPacketSize", &max_packet_size),
    CF_UINT("MaxCommentSize", &max_comment_size),
    CF_LIST("Zone", &zone_list, &zone_config),
    CF_STRING("Database", &database_name),
    CF_STRING("TempKeyFile", &temp_key_file),
    CF_STRING("LogStream", &log_stream_name),
    CF_END
  }
};

static const struct opt_section options = {
  OPT_ITEMS {
    OPT_HELP("A sub-authentication daemon."),
    OPT_HELP("Usage: subauthd [options]"),
    OPT_HELP(""),
    OPT_HELP("Options:"),
    OPT_HELP_OPTION,
    OPT_CONF_OPTIONS,
    OPT_END
  }
};

int main(int argc UNUSED, char **argv)
{
  umask(0077);

  cf_def_file = CONFIG_DIR "/subauthd";
  cf_declare_section("SubauthD", &daemon_config, 0);
  opt_parse(&options, argv+1);

  if (log_stream_name)
    log_configured(log_stream_name);

  auth_init();
  temp_init();
  main_init();
  init_socket();
  init_signals();

  main_loop();
  return 0;
}
